@@ -64,15 +64,15 @@ od_dump_list (ObjectDirectory * od)
 #if 0
 	int i, j, k;
 
-	for (i = 0; i < od->attr_servers._length; i++) {
+	for (i = 0; i < od->attr_servers->_length; i++) {
 		g_print ("IID %s, type %s, location %s\n",
-			 od->attr_servers._buffer[i].iid,
-			 od->attr_servers._buffer[i].server_type,
-			 od->attr_servers._buffer[i].location_info);
-		for (j = 0; j < od->attr_servers._buffer[i].props._length;
+			 od->attr_servers->_buffer[i].iid,
+			 od->attr_servers->_buffer[i].server_type,
+			 od->attr_servers->_buffer[i].location_info);
+		for (j = 0; j < od->attr_servers->_buffer[i].props._length;
 		     j++) {
 			Bonobo_ActivationProperty *prop =
-				&(od->attr_servers._buffer[i].
+				&(od->attr_servers->_buffer[i].
 				  props._buffer[j]);
                         if (strchr (prop->name, '-') != NULL) /* Translated, likely to
                                                          be annoying garbage value */
@@ -184,8 +184,12 @@ update_registry (ObjectDirectory *od, gboolean force_reload)
 #ifdef BONOBO_ACTIVATION_DEBUG
                 g_warning ("Re-load %d %d", must_load, force_reload);
 #endif
+                if (od->attr_servers)
+                        CORBA_free (od->attr_servers);
+                od->attr_servers = CORBA_sequence_Bonobo_ServerInfo__alloc ();
                 bonobo_server_info_load (od->registry_source_directories,
-                                         &od->attr_servers,
+                                         od->attr_servers,
+                                         od->attr_runtime_servers,
                                          &od->by_iid,
                                          bonobo_activation_hostname_get ());
                 od->time_did_stat = od->time_list_changed = time (NULL);
@@ -259,7 +263,7 @@ impl_Bonobo_ObjectDirectory__get_servers (
 
 	retval->_d = (only_if_newer < od->time_list_changed);
 	if (retval->_d) {
-		retval->_u.server_list = od->attr_servers;
+		retval->_u.server_list = *od->attr_servers;
 		CORBA_sequence_set_release (&retval->_u.server_list,
 					    CORBA_FALSE);
 	}
@@ -324,13 +328,13 @@ od_get_active_server (ObjectDirectory    *od,
 
 	retval = CORBA_OBJECT_NIL;
 
-	for (i = 0; i < servers->n_servers; i++)
+	for (i = 0; i < servers->n_servers; i++) {
 		if (Bonobo_ActivationEnvironment_match (
 				&servers->servers [i].environment,
 				environment)) {
 			retval = servers->servers [i].server;
 		}
-
+        }
 	if (retval != CORBA_OBJECT_NIL &&
 	    !CORBA_Object_non_existent (retval, NULL))
 		return CORBA_Object_duplicate (retval, NULL);
@@ -338,6 +342,90 @@ od_get_active_server (ObjectDirectory    *od,
 	return CORBA_OBJECT_NIL;
 }
 
+
+/*
+ * returns (@merged_environment) new environment as result of
+ * merging activation request environment and client registered
+ * environment; the activation supplied environment takes precedence
+ * over the client one
+ */
+static void
+od_merge_client_environment (ObjectDirectory                    *od,
+                             Bonobo_ServerInfo const            *server,
+                             const Bonobo_ActivationEnvironment *environment,
+                             Bonobo_ActivationEnvironment       *merged_environment,
+                             Bonobo_ActivationClient             client)
+{
+        GArray *array;
+        int i, serverinfo_env_idx;
+        const Bonobo_ActivationEnvironment *client_env;
+        const Bonobo_StringList *serverinfo_env = NULL;
+
+        array = g_array_new (FALSE, FALSE, sizeof (Bonobo_ActivationEnvValue));
+
+          /* copy all values from @environment */
+        for (i = 0; i < environment->_length; ++i)
+                g_array_append_val (array, environment->_buffer[i]);
+
+        if (client == CORBA_OBJECT_NIL)
+                goto exit;
+
+        /* scan through server properties */
+        if (!server) goto exit;
+        for (i = 0; i < server->props._length; ++i) {
+                if (strcmp (server->props._buffer[i].name, "bonobo:environment") == 0)
+                {
+                        Bonobo_ActivationPropertyValue const *prop =
+                                &server->props._buffer[i].v;
+                        if (prop->_d == Bonobo_ACTIVATION_P_STRINGV)
+                                serverinfo_env = &prop->_u.value_stringv;
+                        else
+                                g_warning ("bonobo:environment should have type stringv");
+                        break;
+                }
+        }
+        if (!serverinfo_env)
+                goto exit;
+
+        /* do the actual merging */
+        client_env = (const Bonobo_ActivationEnvironment *)
+                g_hash_table_lookup (od->client_envs, client);
+
+        if (!client_env)
+                goto exit;
+
+        for (serverinfo_env_idx = 0;
+             serverinfo_env_idx < serverinfo_env->_length; ++serverinfo_env_idx)
+        {
+                CORBA_char *env = serverinfo_env->_buffer[serverinfo_env_idx];
+                gboolean duplicated_env = FALSE;
+
+                /* check if array already has this environment */
+                for (i = 0; i < environment->_length; ++i) {
+                        if (strcmp (environment->_buffer[i].name, env) == 0) {
+                                duplicated_env = TRUE;
+                                break;
+                        }
+                }
+                if (duplicated_env)
+                        continue;
+
+                /* look for environment in client_env */
+                for (i = 0; i < client_env->_length; ++i) {
+                        if (strcmp (client_env->_buffer[i].name, env) == 0) {
+                                g_array_append_val (array, client_env->_buffer[i]);
+                                break;
+                        }
+                }
+        }
+exit:
+        /* return the resulting environment */
+        merged_environment->_buffer = (Bonobo_ActivationEnvValue *) array->data;
+        merged_environment->_length = merged_environment->_maximum = array->len;
+        g_array_free (array, FALSE);
+}
+
+
 static CORBA_Object
 impl_Bonobo_ObjectDirectory_activate (
 	PortableServer_Servant              servant,
@@ -345,6 +433,7 @@ impl_Bonobo_ObjectDirectory_activate (
 	const Bonobo_ActivationContext      ac,
 	const Bonobo_ActivationEnvironment *environment,
 	const Bonobo_ActivationFlags        flags,
+	Bonobo_ActivationClient             client,
 	CORBA_Context                       ctx,
 	CORBA_Environment                  *ev)
 {
@@ -355,16 +444,23 @@ impl_Bonobo_ObjectDirectory_activate (
 #ifdef BONOBO_ACTIVATION_DEBUG
 	static int                       depth = 0;
 #endif
+        Bonobo_ActivationEnvironment merged_environment;
+
+        od_merge_client_environment (od, (Bonobo_ServerInfo *)
+                                     g_hash_table_lookup (od->by_iid, iid),
+                                     environment, &merged_environment, client);
 
 	retval = CORBA_OBJECT_NIL;
 
         update_registry (od, FALSE);
 
         if (!(flags & Bonobo_ACTIVATION_FLAG_PRIVATE)) {
-                retval = od_get_active_server (od, iid, environment);
+                retval = od_get_active_server (od, iid, &merged_environment);
 
-                if (retval != CORBA_OBJECT_NIL)
+                if (retval != CORBA_OBJECT_NIL) {
+                        g_free (merged_environment._buffer);
                         return retval;
+                }
         }
 
 	if (flags & Bonobo_ACTIVATION_FLAG_EXISTING_ONLY) {
@@ -389,7 +485,7 @@ impl_Bonobo_ObjectDirectory_activate (
 
 	if (si) {
 		retval = od_server_activate (
-				si, &ai, BONOBO_OBJREF (od), environment, ev);
+				si, &ai, BONOBO_OBJREF (od), &merged_environment, client, ev);
 
                 /* If we failed to activate - it may be because our
                  * request re-entered _during_ the activation
@@ -407,7 +503,7 @@ impl_Bonobo_ObjectDirectory_activate (
                 /* FIXME: this path is theoretically redundant now */
                 if (ev->_major != CORBA_NO_EXCEPTION ||
                     retval == CORBA_OBJECT_NIL) {
-                        retval = od_get_active_server (od, iid, environment);
+                        retval = od_get_active_server (od, iid, &merged_environment);
 
                         if (retval != CORBA_OBJECT_NIL)
                                 CORBA_exception_free (ev);
@@ -423,6 +519,7 @@ impl_Bonobo_ObjectDirectory_activate (
                 depth--;
         }
 #endif
+        g_free (merged_environment._buffer);
 
 	return retval;
 }
@@ -585,7 +682,8 @@ add_active_server (ObjectDirectory    *od,
                                 NULL);
                 }
         } else
-                g_assert (!strcmp (iid, NAMING_CONTEXT_IID));
+                g_assert (!strcmp (iid, NAMING_CONTEXT_IID) ||
+                          !strcmp(iid, EVENT_SOURCE_IID));
 
 	servers = g_hash_table_lookup (od->active_server_lists, iid);
 	if (!servers) {
@@ -647,7 +745,8 @@ remove_active_server (ObjectDirectory *od,
 	int               i;
 
 	servers = g_hash_table_lookup (od->active_server_lists, iid);
-        g_return_val_if_fail (servers != NULL, FALSE);
+        if (!servers)
+                return FALSE;
 
 	for (i = 0; i < servers->n_servers; i++)
 		if (CORBA_Object_is_equivalent (
@@ -667,6 +766,137 @@ remove_active_server (ObjectDirectory *od,
 	return removed;
 }
 
+  /* Parse server description and register it, replacing older
+   * definition if necessary.  Returns the regsitered ServerInfo */
+static Bonobo_ServerInfo const *
+od_register_runtime_server_info (ObjectDirectory  *od,
+                                 const char       *iid,
+                                 const CORBA_char *description)
+{
+        Bonobo_ServerInfo *old_serverinfo, *new_serverinfo;
+        GSList *parsed_serverinfo = NULL, *l;
+        int     i;
+
+        old_serverinfo = (Bonobo_ServerInfo *) g_hash_table_lookup (od->by_iid, iid);
+        if (!(*description)) /* empty description? */
+                return old_serverinfo;
+
+          /* parse description */
+         bonobo_parse_server_info_memory (description, &parsed_serverinfo,
+                                          bonobo_activation_hostname_get ());
+
+           /* check for zero entries */
+         if (!parsed_serverinfo)
+                 return NULL;
+           /* check for more than one entry */
+         if (parsed_serverinfo->next) {
+                 g_warning ("More than one <oaf_server> specified, ignoring all");
+                 for (l = parsed_serverinfo; l; l = l->next) {
+                         Bonobo_ServerInfo__freekids (l->data, NULL);
+                         g_free (l->data);
+                 }
+                 g_slist_free (parsed_serverinfo);
+                 return NULL;
+         }
+         new_serverinfo = (Bonobo_ServerInfo *) parsed_serverinfo->data;
+         g_slist_free (parsed_serverinfo);
+
+         if (old_serverinfo) {
+                 Bonobo_ServerInfo *old_runtime_serverinfo;
+                   /* find and replace new serverinfo in attr_runtime_servers */
+                 for (i = 0; i < od->attr_runtime_servers->len; i++) {
+                         old_runtime_serverinfo = g_ptr_array_index (od->attr_runtime_servers, i);
+                         if (strcmp (old_runtime_serverinfo->iid, iid) == 0) {
+                                 Bonobo_ServerInfo__freekids (old_runtime_serverinfo, NULL);
+                                 g_free (old_runtime_serverinfo);
+                                 g_ptr_array_index (od->attr_runtime_servers, i) = new_serverinfo;
+                                 break;
+                         }
+                 }
+                 if (i == od->attr_runtime_servers->len) {
+                         g_warning ("Inconsistency between od->attr_servers "
+                                    "and od->attr_runtime_servers");
+                           /* something must be wrong, but we try to remedy as we can */
+                         g_ptr_array_add (od->attr_runtime_servers, new_serverinfo);
+                 }
+                   /* replace old for new serverinfo in attr_servers */
+                 g_hash_table_remove (od->by_iid, old_serverinfo->iid);
+                 Bonobo_ServerInfo__freekids (old_serverinfo, NULL);
+                 Bonobo_ServerInfo_copy (old_serverinfo, new_serverinfo);
+                 g_hash_table_insert (od->by_iid, old_serverinfo->iid, old_serverinfo);
+         } else {
+                 g_ptr_array_add (od->attr_runtime_servers, new_serverinfo);
+                 ORBit_sequence_append (od->attr_servers, new_serverinfo);
+                   /* rebuild od->by_iid hash table, because
+                    * ORBit_sequence_append reallocs _buffer, and that
+                    * sometimes changes the addresses of the
+                    * serverinfo items */
+                 g_hash_table_destroy (od->by_iid);
+                 od->by_iid = g_hash_table_new (g_str_hash, g_str_equal);
+                 for (i = 0; i < od->attr_servers->_length; ++i)
+                         g_hash_table_insert (od->by_iid,
+                                              od->attr_servers->_buffer[i].iid,
+                                              od->attr_servers->_buffer + i);
+         }
+         od->time_list_changed = time (NULL);
+         activation_clients_cache_notify ();
+         return old_serverinfo;
+}
+
+static Bonobo_RegistrationResult
+impl_Bonobo_ObjectDirectory_register_new_full (
+	PortableServer_Servant              servant,
+	const CORBA_char                   *iid,
+	const Bonobo_ActivationEnvironment *environment,
+	const CORBA_Object                  obj,
+        Bonobo_RegistrationFlags            flags,
+        const CORBA_char                   *description,
+        CORBA_Object                       *existing,
+        Bonobo_ActivationClient             client,
+	CORBA_Environment                  *ev)
+{
+	ObjectDirectory              *od = OBJECT_DIRECTORY (servant);
+	CORBA_Object                  oldobj;
+        Bonobo_ActivationEnvironment  merged_environment;
+        Bonobo_ServerInfo const      *serverinfo;
+
+	oldobj = od_get_active_server (od, iid, environment);
+        *existing = oldobj;
+
+        serverinfo = od_register_runtime_server_info (od, iid, description);
+        od_merge_client_environment (od, serverinfo, environment,
+                                     &merged_environment, client);
+
+	oldobj = od_get_active_server (od, iid, &merged_environment);
+	if (oldobj != CORBA_OBJECT_NIL) {
+		if (!CORBA_Object_non_existent (oldobj, ev)) {
+                        g_free (merged_environment._buffer);
+			return Bonobo_ACTIVATION_REG_ALREADY_ACTIVE;
+                }
+	}
+
+        if (!serverinfo) {
+                if (!(flags&Bonobo_REGISTRATION_FLAG_NO_SERVERINFO)) {
+                        g_free (merged_environment._buffer);
+                        return Bonobo_ACTIVATION_REG_NOT_LISTED;
+                }
+        }
+
+#ifdef BONOBO_ACTIVATION_DEBUG
+        g_warning ("Server register. '%s' : %p", iid, obj);
+#endif
+
+        add_active_server (od, iid, &merged_environment, obj);
+        g_free (merged_environment._buffer);
+	
+	bonobo_event_source_notify_listeners
+                (od->event_source,
+                 "Bonobo/ObjectDirectory:activation:register",
+                 NULL, ev);
+
+	return Bonobo_ACTIVATION_REG_SUCCESS;
+}
+
 static Bonobo_RegistrationResult
 impl_Bonobo_ObjectDirectory_register_new (
 	PortableServer_Servant              servant,
@@ -678,29 +908,9 @@ impl_Bonobo_ObjectDirectory_register_new (
         CORBA_Object                       *existing,
 	CORBA_Environment                  *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-	CORBA_Object     oldobj;
-
-	oldobj = od_get_active_server (od, iid, environment);
-        *existing = oldobj;
-
-	if (oldobj != CORBA_OBJECT_NIL) {
-		if (!CORBA_Object_non_existent (oldobj, ev))
-			return Bonobo_ACTIVATION_REG_ALREADY_ACTIVE;
-	}
-
-        if (!g_hash_table_lookup (od->by_iid, iid)) {
-                if (!(flags&Bonobo_REGISTRATION_FLAG_NO_SERVERINFO))
-                        return Bonobo_ACTIVATION_REG_NOT_LISTED;
-        }
-
-#ifdef BONOBO_ACTIVATION_DEBUG
-        g_warning ("Server register. '%s' : %p", iid, obj);
-#endif
-
-        add_active_server (od, iid, environment, obj);
-
-	return Bonobo_ACTIVATION_REG_SUCCESS;
+        return impl_Bonobo_ObjectDirectory_register_new_full
+                (servant, iid, environment, obj, flags,
+                 description, existing, CORBA_OBJECT_NIL, ev);
 }
 
 static void
@@ -716,6 +926,11 @@ impl_Bonobo_ObjectDirectory_unregister (
                 CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
                                      ex_Bonobo_ObjectDirectory_NotRegistered,
                                      NULL);
+	else 
+                bonobo_event_source_notify_listeners
+                        (od->event_source,
+                         "Bonobo/ObjectDirectory:activation:unregister",
+                         NULL, ev);
 }
 
 static Bonobo_DynamicPathLoadResult 
@@ -831,6 +1046,60 @@ impl_Bonobo_ObjectDirectory_remove_path(
 	return Bonobo_DYNAMIC_LOAD_SUCCESS;
 }
 
+
+static void
+client_cnx_broken (ORBitConnection *cnx,
+                   const Bonobo_ActivationClient  client)
+{
+        ObjectDirectory *od = main_dir;
+        if (!od) /* shutting down */
+                return;
+        g_hash_table_remove (od->client_envs, client);
+}
+
+static void
+impl_Bonobo_ObjectDirectory_addClientEnv (
+        PortableServer_Servant         servant,
+        const Bonobo_ActivationClient  client,
+        const Bonobo_StringList       *client_env,
+        CORBA_Environment             *ev)
+{
+        Bonobo_ActivationEnvironment *env;
+	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+        int i;
+        
+        env = Bonobo_ActivationEnvironment__alloc ();
+        env->_length  = env->_maximum = client_env->_length;
+        env->_buffer  = Bonobo_ActivationEnvironment_allocbuf (env->_length);
+        env->_release = CORBA_TRUE;
+
+        for (i = 0; i < client_env->_length; ++i)
+        {
+                const char *keyval = client_env->_buffer[i];
+                const char *equals = strchr (keyval, '=');
+                guint       keylen;
+
+                if (!equals) {
+                        g_warning ("Duff env. var '%s'", keyval);
+                        continue;
+                }
+
+                keylen = (guint) (equals - keyval);
+
+                env->_buffer[i].name = CORBA_string_alloc (keylen + 1);
+                strncpy (env->_buffer[i].name, keyval, keylen);
+                env->_buffer[i].name[keylen] = 0;
+                env->_buffer[i].value = CORBA_string_dup (equals + 1);
+                env->_buffer[i].flags = 0;
+        }
+
+        g_hash_table_insert (od->client_envs, client, env);
+
+        ORBit_small_listen_for_broken (client, G_CALLBACK (client_cnx_broken),
+                                       (gpointer) client);
+}
+
+
 Bonobo_ObjectDirectory
 bonobo_object_directory_get (void)
 {
@@ -840,6 +1109,15 @@ bonobo_object_directory_get (void)
                 return BONOBO_OBJREF (main_dir);
 }
 
+Bonobo_EventSource
+bonobo_object_directory_event_source_get (void)
+{
+     if (!main_dir)
+    	      return CORBA_OBJECT_NIL;
+     else
+              return BONOBO_OBJREF (main_dir->event_source);
+}
+
 void
 bonobo_object_directory_init (PortableServer_POA poa,
                               const char        *registry_path,
@@ -901,9 +1179,8 @@ bonobo_object_directory_re_check_fn (const Bonobo_ActivationEnvironment *environ
 void
 bonobo_object_directory_reload (void)
 {
-#ifdef BONOBO_ACTIVATION_DEBUG
-        g_print ("reloading our object directory!\n");
-#endif
+        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reloading our object directory!");
+
         update_registry (main_dir, TRUE);
 }
 
@@ -919,6 +1196,11 @@ object_directory_finalize (GObject *object)
 
         g_strfreev (od->registry_source_directories);
 
+        if (od->client_envs) {
+                g_hash_table_destroy (od->client_envs);
+                od->client_envs = NULL;
+        }
+
         parent_class->finalize (object);
 }
 
@@ -935,9 +1217,13 @@ object_directory_class_init (ObjectDirectoryClass *klass)
         epv->get_active_servers  = impl_Bonobo_ObjectDirectory_get_active_servers;
 	epv->activate            = impl_Bonobo_ObjectDirectory_activate;
 	epv->register_new        = impl_Bonobo_ObjectDirectory_register_new;
+	epv->register_new_full   = impl_Bonobo_ObjectDirectory_register_new_full;
 	epv->unregister          = impl_Bonobo_ObjectDirectory_unregister;
 	epv->dynamic_add_path    = impl_Bonobo_ObjectDirectory_add_path;
 	epv->dynamic_remove_path = impl_Bonobo_ObjectDirectory_remove_path;
+        epv->addClientEnv        = impl_Bonobo_ObjectDirectory_addClientEnv;
+
+	epv->register_new        = impl_Bonobo_ObjectDirectory_register_new;
 }
 
 static void
@@ -953,6 +1239,13 @@ object_directory_init (ObjectDirectory *od)
                 g_hash_table_new_full (g_str_hash, g_str_equal,
                                        g_free, active_server_list_free);
         od->no_servers_timeout = 0;
+
+        od->attr_runtime_servers = g_ptr_array_new ();
+	
+	od->event_source = bonobo_event_source_new ();
+        od->client_envs = g_hash_table_new_full
+                (NULL, NULL, NULL,
+                 (GDestroyNotify) CORBA_free);
 }
 
 BONOBO_TYPE_FUNC_FULL (ObjectDirectory,
