@@ -51,7 +51,8 @@ bonobo_activation_timeout_reg_check_set (gboolean on)
 gboolean
 bonobo_activation_timeout_reg_check (gpointer data)
 {
-        if (!check_registration)
+        if (!check_registration ||
+            g_getenv ("BONOBO_ACTIVATION_NO_TIMEOUT"))
                 return FALSE;
 
         if (need_ior_printout) {
@@ -301,7 +302,7 @@ bonobo_activation_register_active_server_ext (const char               *iid,
                          existing, bonobo_activation_client_get (), &ev);
 
                 if (ev._major == CORBA_SYSTEM_EXCEPTION &&
-                    !strcmp (ev._id, ex_CORBA_NO_IMPLEMENT)) /* fall-back */
+                    !strcmp (ev._id, ex_CORBA_BAD_OPERATION)) /* fall-back */
                         retval = Bonobo_ObjectDirectory_register_new
                                 (od, iid, reg_env ? &environment : &global_reg_env,
                                  obj, flags, description ? description : "",
